#include <stdio.h>
#include <tchar.h>
#include <windows.h>
#include <winioctl.h>
#include <conio.h>

/*
 * NTRawrite.c by Blake Ramsdell <bcr@users.sourceforge.net>
 *
 * Created because Rawrite/Rawrite3 is using Turbo C calls from 1990.
 *
 * Use this as you will.  Just leave my name on it and tell me what you did.
 *
 */

#define APPLICATION_NAME    TEXT("NTRawrite")
#define APPLICATION_VERSION TEXT("1.0.1")

static void
LastErrorPerror(
    LPCTSTR pszPrefixText
    )
{
    LPVOID  lpMessageBuffer = NULL;
    DWORD   dwLastError     = 0;

    dwLastError = GetLastError();

    (void) FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMessageBuffer,
        0,
        NULL
        );

    if (lpMessageBuffer != NULL)
    {
        /* Display it */

        _ftprintf(
            stderr,
            TEXT("%s: %s"),
            pszPrefixText,
            lpMessageBuffer
            );

        /* Free the buffer allocated by the system */

        LocalFree( lpMessageBuffer );
        lpMessageBuffer = NULL;
    }

    /* Put the error back */

    SetLastError(dwLastError);
}

static BOOL
DoCopyFileWithStatus(
    HANDLE  hSourceFile,
    DWORD   dwSourceFileLength,
    HANDLE  hDestinationFile,
    DWORD   dwBlockSize
    )
{
    unsigned char   buffer[4096];
    BOOL            bReturnCode = TRUE;
    DWORD           dwTotalBytesCopied = 0;
    unsigned int    nLastDotOutput = 0;
    DWORD           dwBufferSizeToUse = 0;

    /* Determine what the number of bytes per read should be */

    dwBufferSizeToUse = (sizeof(buffer) / dwBlockSize) * dwBlockSize;

    /* If not supplied, get the file length */

    if (dwSourceFileLength == 0)
    {
        BY_HANDLE_FILE_INFORMATION  ourFileInformation;

        memset(&ourFileInformation, 0, sizeof(ourFileInformation));

        bReturnCode = GetFileInformationByHandle(
            hSourceFile,
            &ourFileInformation
            );
        if ((!bReturnCode) && (GetLastError() != ERROR_INVALID_FUNCTION))
        {
            LastErrorPerror(APPLICATION_NAME);
            return bReturnCode;
        }

        dwSourceFileLength = ourFileInformation.nFileSizeLow;
    }

    /* If we have a file length, then show better status. */

    if (dwSourceFileLength > 0)
    {
        /* Fix the source file length to be a multiple of the block size */

        if ((dwSourceFileLength % dwBlockSize) > 0)
        {
            dwSourceFileLength += (dwBlockSize - (dwSourceFileLength % dwBlockSize));
        }

        /* 40 dots in here -- I only figured out afterwards that this was a good
           number (1.44MB floppies are 80 tracks) */

        _tprintf(TEXT("........................................]\r"));
    }

    while (1)
    {
        DWORD   dwBytesRead = 0;
        DWORD   dwBytesWritten = 0;

        bReturnCode = ReadFile(
            hSourceFile,
            buffer,
            dwBufferSizeToUse,
            &dwBytesRead,
            NULL
            );
        if (!bReturnCode)
        {
            _tprintf(TEXT("\n"));
            if (GetLastError() == ERROR_INVALID_PARAMETER)
            {
                /* All done -- this is what the floppy returns when we get to
                   the end.  Helpful, isn't it? */

                bReturnCode = TRUE;
                break;
            }
            LastErrorPerror(APPLICATION_NAME);
            break;
        }

        if (dwBytesRead == 0)
        {
            /* All done */

            break;
        }

        /* Modify dwBytesRead in case it isn't a multiple of the block size */

        if ((dwBytesRead % dwBlockSize) > 0)
        {
            DWORD   dwDifference = (dwBlockSize - (dwBytesRead % dwBlockSize));

            memset(buffer + dwBytesRead, 0, dwDifference);
            dwBytesRead += dwDifference;
        }

        bReturnCode = WriteFile(
            hDestinationFile,
            buffer,
            dwBytesRead,
            &dwBytesWritten,
            NULL
            );
        if (!bReturnCode)
        {
            _tprintf(TEXT("\n"));
            LastErrorPerror(APPLICATION_NAME);
            break;
        }

        dwTotalBytesCopied += dwBytesRead;

        /* Update status */

        if (dwSourceFileLength > 0)
        {
            UINT    nThisDotOutput = 0;

            nThisDotOutput = ((dwTotalBytesCopied * 40) / dwSourceFileLength);
            if (nLastDotOutput < nThisDotOutput)
            {
                _puttchar(TEXT('*'));
                nLastDotOutput = nThisDotOutput;
            }
        }
        else
        {
            _tprintf(TEXT("\r%lu bytes copied."), dwTotalBytesCopied);
        }
    }

    _tprintf(TEXT("\n"));
    return bReturnCode;
}

static BOOL
DoVerifyFileWithStatus(
    HANDLE  hSourceFile,
    HANDLE  hDestinationFile,
    DWORD   dwBlockSize
    )
{
    unsigned char   buffer[4096];
    unsigned char   buffer2[4096];
    BOOL            bReturnCode = TRUE;
    DWORD           dwSourceFileLength = 0;
    DWORD           dwTotalBytesVerified = 0;
    unsigned int    nLastDotOutput = 0;
    DWORD           dwBufferSizeToUse = 0;

    /* Determine what the number of bytes per read should be */

    dwBufferSizeToUse = (sizeof(buffer) / dwBlockSize) * dwBlockSize;

    /* Get the file length */

    {
        BY_HANDLE_FILE_INFORMATION  ourFileInformation;

        memset(&ourFileInformation, 0, sizeof(ourFileInformation));

        bReturnCode = GetFileInformationByHandle(
            hSourceFile,
            &ourFileInformation
            );
        if (!bReturnCode)
        {
            LastErrorPerror(APPLICATION_NAME);
            return bReturnCode;
        }

        dwSourceFileLength = ourFileInformation.nFileSizeLow;
    }

    /* Fix the source file length to be a multiple of the block size */

    if ((dwSourceFileLength % dwBlockSize) > 0)
    {
        dwSourceFileLength += (dwBlockSize - (dwSourceFileLength % dwBlockSize));
    }

    /* 40 dots in here -- I only figured out afterwards that this was a good
       number (1.44MB floppies are 80 tracks) */

    _tprintf(TEXT("........................................]\r"));

    while (1)
    {
        DWORD   dwBytesRead = 0;
        DWORD   dwBytesWritten = 0;

        bReturnCode = ReadFile(
            hSourceFile,
            buffer,
            dwBufferSizeToUse,
            &dwBytesRead,
            NULL
            );
        if (!bReturnCode)
        {
            _tprintf(TEXT("\n"));
            LastErrorPerror(APPLICATION_NAME);
            break;
        }

        if (dwBytesRead == 0)
        {
            /* All done */

            break;
        }

        /* Modify dwBytesRead in case it isn't a multiple of the block size */

        if ((dwBytesRead % dwBlockSize) > 0)
        {
            DWORD   dwDifference = (dwBlockSize - (dwBytesRead % dwBlockSize));

            memset(buffer + dwBytesRead, 0, dwDifference);
            dwBytesRead += dwDifference;
        }

        bReturnCode = ReadFile(
            hDestinationFile,
            buffer2,
            dwBytesRead,
            &dwBytesRead,
            NULL
            );
        if (!bReturnCode)
        {
            _tprintf(TEXT("\n"));
            LastErrorPerror(APPLICATION_NAME);
            break;
        }

        if (memcmp(buffer, buffer2, dwBytesRead) != 0)
        {
            _tprintf(TEXT("\n"));
            _ftprintf(stderr, TEXT("%s: Failed to verify -- try another diskette or reformatting."), APPLICATION_NAME);
            bReturnCode = FALSE;
            break;
        }

        dwTotalBytesVerified += dwBytesRead;

        /* Update status */

        {
            UINT    nThisDotOutput = 0;

            nThisDotOutput = ((dwTotalBytesVerified * 40) / dwSourceFileLength);
            if (nLastDotOutput < nThisDotOutput)
            {
                _puttchar(TEXT('*'));
                nLastDotOutput = nThisDotOutput;
            }
        }
    }

    _tprintf(TEXT("\n"));
    return bReturnCode;
}

static void usage(void)
{
    _tprintf(TEXT("Creates a diskette from a diskette image file.\n\n"));
    _tprintf(TEXT("usage: %s [--noverify] [--reverse] [-n] [-h] [-f image_file] [-d drive]\n\n"), APPLICATION_NAME);
    _tprintf(TEXT("       --noverify     Skips verification step.\n"));
    _tprintf(TEXT("       --reverse      Saves diskette to image file instead.\n"));
    _tprintf(TEXT("       -n             Don\'t wait for the user to insert a diskette.\n"));
    _tprintf(TEXT("       -f image_file  Disk image file to place on diskette.\n"));
    _tprintf(TEXT("       -d drive       Drive specifier to put diskette image on.\n"));
    _tprintf(TEXT("       -h             Display usage.\n"));
}

int main(int argc, char** argv)
{
    HANDLE  hDisketteFile       = INVALID_HANDLE_VALUE;
    HANDLE  hImageFile          = INVALID_HANDLE_VALUE;
    LPCTSTR pszImagePathname    = NULL;
    LPCTSTR pszDriveArgument    = NULL;
    BOOL    bVerifyCopy         = TRUE;
    BOOL    bReverseDirection   = FALSE;
    DWORD   dwFloppySize        = 0;
    DWORD   dwBlockSize         = 0;
    DWORD   dwImageFileSize     = 0;
    BOOL    bReturnCode         = FALSE;
    BOOL    bWaitForInsertion   = TRUE;

    TCHAR   szDiskPathname[_MAX_PATH];
    TCHAR   szImagePathname[_MAX_PATH];
    TCHAR   szDriveArgument[_MAX_PATH];

    _tprintf(TEXT("%s v%s by Blake Ramsdell <bcr@users.sourceforge.net>\n\n"), APPLICATION_NAME, APPLICATION_VERSION);

    /* Figure out the arguments handed in */

    {
        int nCounter = 0;

        for (nCounter = 1;nCounter < argc;++nCounter)
        {
            if (_tcsicmp(argv[nCounter], TEXT("--noverify")) == 0)
            {
                bVerifyCopy = FALSE;
            }
            else if (_tcsicmp(argv[nCounter], TEXT("--reverse")) == 0)
            {
                bReverseDirection = TRUE;
            }
            else if (_tcsicmp(argv[nCounter], TEXT("-n")) == 0)
            {
                bWaitForInsertion = FALSE;
            }
            else if (_tcsicmp(argv[nCounter], TEXT("-h")) == 0)
            {
                usage();
                return 0;
            }
            else if (_tcsicmp(argv[nCounter], TEXT("-f")) == 0)
            {
                if ((nCounter + 1) < argc)
                {
                    pszImagePathname = argv[nCounter + 1];
                    ++nCounter;
                }
                else
                {
                    _ftprintf(stderr, TEXT("%s: Must specify an image filename after -f.\n"), APPLICATION_NAME);
                    usage();
                    return 1;
                }
            }
            else if (_tcsicmp(argv[nCounter], TEXT("-d")) == 0)
            {
                if ((nCounter + 1) < argc)
                {
                    pszDriveArgument = argv[nCounter + 1];
                    ++nCounter;
                }
                else
                {
                    _ftprintf(stderr, TEXT("%s: Must specify a drive after -d.\n"), APPLICATION_NAME);
                    usage();
                    return 1;
                }
            }
            else
            {
                usage();
                return 1;
            }
        }
    }

    /* Get the image pathname from the user if not provided */

    if (pszImagePathname == NULL)
    {
        LPTSTR  pszReturnValue = NULL;
        LPTSTR  pszLF = NULL;

        memset(szImagePathname, 0, sizeof(szImagePathname));

        _tprintf(TEXT("Please type the image pathname: "));
        pszReturnValue = _fgetts(
            szImagePathname,
            sizeof(szImagePathname) / sizeof(szImagePathname[0]),
            stdin
            );
        if (pszReturnValue == NULL)
        {
            LastErrorPerror(APPLICATION_NAME);
            return 1;
        }

        /* Trim off the trailing LF */

        pszLF = _tcschr(szImagePathname, TEXT('\n'));
        if (pszLF != NULL)
        {
            *pszLF = TEXT('\0');
        }

        pszImagePathname = szImagePathname;
    }

    /* Get the drive from the user if not provided */

    if (pszDriveArgument == NULL)
    {
        LPTSTR  pszReturnValue = NULL;
        LPTSTR  pszLF = NULL;

        memset(szDriveArgument, 0, sizeof(szDriveArgument));

        _tprintf(TEXT("Please type the diskette drive: "));
        pszReturnValue = _fgetts(
            szDriveArgument,
            sizeof(szDriveArgument) / sizeof(szDriveArgument[0]),
            stdin
            );
        if (pszReturnValue == NULL)
        {
            LastErrorPerror(APPLICATION_NAME);
            return 1;
        }

        /* Trim off the trailing LF */

        pszLF = _tcschr(szDriveArgument, TEXT('\n'));
        if (pszLF != NULL)
        {
            *pszLF = TEXT('\0');
        }

        pszDriveArgument = szDriveArgument;
    }

    if ((pszImagePathname == NULL) || (pszDriveArgument == NULL))
    {
        usage();
        return 1;
    }

    if (bWaitForInsertion)
    {
        _tprintf(TEXT("Please insert a diskette and press any key."));
        _getche();
        _tprintf(TEXT("\n"));
    }

    /* Make the pathname for the disk drive */

    _stprintf(szDiskPathname, TEXT("\\\\.\\%c:"), _totupper(_tcsnextc(pszDriveArgument)));

    /* Open the diskette */

    hDisketteFile = CreateFile(
        szDiskPathname,
        (bReverseDirection) ? GENERIC_READ : GENERIC_WRITE,
        0,
        NULL,
        OPEN_EXISTING,
        0,
        NULL
        );
    if (hDisketteFile == INVALID_HANDLE_VALUE)
    {
        LastErrorPerror(APPLICATION_NAME);
        return 1;
    }

	/* Figure out the floppy geometry */

	{
        DISK_GEOMETRY	ourDiskGeometry;
        DWORD           dwBytesReturned = 0;

        memset(&ourDiskGeometry, 0, sizeof(ourDiskGeometry));

        bReturnCode = DeviceIoControl(hDisketteFile,
            IOCTL_DISK_GET_DRIVE_GEOMETRY,
            NULL,
            0,
            &ourDiskGeometry,
            sizeof(ourDiskGeometry),
            &dwBytesReturned,
            (LPOVERLAPPED) NULL
            );
        if (!bReturnCode)
        {
            LastErrorPerror(APPLICATION_NAME);
            return 1;
        }

        dwFloppySize =  ourDiskGeometry.Cylinders.LowPart *
                        ourDiskGeometry.TracksPerCylinder *
                        ourDiskGeometry.SectorsPerTrack *
                        ourDiskGeometry.BytesPerSector;

        dwBlockSize = ourDiskGeometry.BytesPerSector;
	}

    /* Open the image file */

    hImageFile = CreateFile(
        pszImagePathname,
        (bReverseDirection) ? GENERIC_WRITE : GENERIC_READ,
        0,
        NULL,
        (bReverseDirection) ? CREATE_ALWAYS : OPEN_EXISTING,
        0,
        NULL
        );
    if (hImageFile == INVALID_HANDLE_VALUE)
    {
        LastErrorPerror(APPLICATION_NAME);
        (void) CloseHandle(hDisketteFile);
        hDisketteFile = INVALID_HANDLE_VALUE;
        return 2;
    }

    /* Get the file length */

    {
        BY_HANDLE_FILE_INFORMATION  ourFileInformation;

        memset(&ourFileInformation, 0, sizeof(ourFileInformation));

        bReturnCode = GetFileInformationByHandle(
            hImageFile,
            &ourFileInformation
            );
        if (!bReturnCode)
        {
            LastErrorPerror(APPLICATION_NAME);
            return 1;
        }

        dwImageFileSize = ourFileInformation.nFileSizeLow;
    }

    /* Make sure that the floppy will fit the data from the image file */

    if ((!bReverseDirection) && (dwFloppySize < dwImageFileSize))
    {
        _ftprintf(
            stderr,
            TEXT("%s: Image size mismatch -- floppy can hold %lu bytes, but image file is %lu bytes long.\n"),
            APPLICATION_NAME,
            dwFloppySize,
            dwImageFileSize
            );
        return 1;
    }

    if (bReverseDirection)
    {
        _tprintf(TEXT("Copying floppy drive %c to image file %s.\n"), _totupper(_tcsnextc(pszDriveArgument)), pszImagePathname);
    }
    else
    {
        _tprintf(TEXT("Copying image file %s to floppy drive %c.\n"), pszImagePathname, _totupper(_tcsnextc(pszDriveArgument)));
    }

    /* Now copy it */

    if (!DoCopyFileWithStatus(
        (bReverseDirection) ? hDisketteFile : hImageFile,
        (bReverseDirection) ? dwFloppySize : dwImageFileSize,
        (bReverseDirection) ? hImageFile : hDisketteFile,
        dwBlockSize
        ))
    {
        (void) CloseHandle(hImageFile);
        hImageFile = INVALID_HANDLE_VALUE;
        (void) CloseHandle(hDisketteFile);
        hDisketteFile = INVALID_HANDLE_VALUE;
        return 3;
    }

    (void) CloseHandle(hImageFile);
    hImageFile = INVALID_HANDLE_VALUE;
    (void) CloseHandle(hDisketteFile);
    hDisketteFile = INVALID_HANDLE_VALUE;

    /* If we need to verify the copy, then do it */

    if (bVerifyCopy)
    {
        /* Open the diskette */

        hDisketteFile = CreateFile(
            szDiskPathname,
            GENERIC_READ,
            0,
            NULL,
            OPEN_EXISTING,
            0,
            NULL
            );
        if (hDisketteFile == INVALID_HANDLE_VALUE)
        {
            LastErrorPerror(APPLICATION_NAME);
            return 4;
        }

        /* Open the image file for reading */

        hImageFile = CreateFile(
            pszImagePathname,
            GENERIC_READ,
            0,
            NULL,
            OPEN_EXISTING,
            0,
            NULL
            );
        if (hImageFile == INVALID_HANDLE_VALUE)
        {
            LastErrorPerror(APPLICATION_NAME);
            (void) CloseHandle(hDisketteFile);
            hDisketteFile = INVALID_HANDLE_VALUE;
            return 5;
        }

        _tprintf(TEXT("Verifying image file %s with floppy drive %c.\n"), pszImagePathname, _totupper(_tcsnextc(pszDriveArgument)));

        /* Now verify it */

        if (!DoVerifyFileWithStatus(
            hImageFile,
            hDisketteFile,
            dwBlockSize
            ))
        {
            (void) CloseHandle(hImageFile);
            hImageFile = INVALID_HANDLE_VALUE;
            (void) CloseHandle(hDisketteFile);
            hDisketteFile = INVALID_HANDLE_VALUE;
            return 6;
        }
    }

    _tprintf(TEXT("Completed!\n"));

    return 0;
}
